from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from magicgui.widgets import Widget

import magicgui

import btrack.napari.constants


def _create_hypotheses_widgets() -> list[Widget]:
    """Create widgets for selecting which hypotheses to generate."""

    hypotheses = btrack.napari.constants.HYPOTHESES
    tooltips = [
        "Hypothesis that a tracklet is a false positive detection. Always required.",
        "Hypothesis that a tracklet starts at the beginning of the movie or edge of the field of view.",  # noqa: E501
        "Hypothesis that a tracklet ends at the end of the movie or edge of the field of view.",  # noqa: E501
        "Hypothesis that two tracklets should be linked together.",
        "Hypothesis that a tracklet can split into two daughter tracklets.",
        "Hypothesis that a tracklet terminates without leaving the field of view.",
        "Hypothesis that two tracklets merge into one tracklet.",
    ]

    hypotheses_widgets = []
    for hypothesis, tooltip in zip(hypotheses, tooltips):
        widget = magicgui.widgets.create_widget(
            value=True,
            name=hypothesis,
            label=hypothesis,
            widget_type="CheckBox",
            options={"tooltip": tooltip},
        )
        hypotheses_widgets.append(widget)

    # P_FP is always required
    P_FP_hypothesis = hypotheses_widgets[0]
    P_FP_hypothesis.enabled = False

    # P_merge should be disabled by default
    P_merge_hypothesis = hypotheses_widgets[-1]
    P_merge_hypothesis.value = False

    return hypotheses_widgets


def _create_scaling_factor_widgets() -> list[Widget]:
    """Create widgets for setting the scaling factors of the HypothesisModel"""

    widget_values = [5.0, 3.0, 10.0, 50.0]
    names = [
        "lambda_time",
        "lambda_dist",
        "lambda_link",
        "lambda_branch",
    ]
    labels = [
        "λ time",
        "λ distance",
        "λ linking",
        "λ branching",
    ]
    tooltips = [
        "Scaling factor for the influence of time when determining initialization or termination hypotheses.",  # noqa: E501
        "Scaling factor for the influence of distance at the border when determining initialization or termination hypotheses.",  # noqa: E501
        "Scaling factor for the influence of track-to-track distance on linking probability.",  # noqa: E501
        "Scaling factor for the influence of cell state and position on division (mitosis/branching) probability.",  # noqa: E501
    ]

    scaling_factor_widgets = []
    for value, name, label, tooltip in zip(
        widget_values, names, labels, tooltips
    ):
        widget = magicgui.widgets.create_widget(
            value=value,
            name=name,
            label=label,
            widget_type="FloatSpinBox",
            options={"tooltip": tooltip},
        )
        scaling_factor_widgets.append(widget)

    return scaling_factor_widgets


def _create_threshold_widgets() -> list[Widget]:
    """Create widgets for setting thresholds for the HypothesisModel"""

    distance_threshold_tooltip = (
        "A threshold distance from the edge of the field of view to add an "
        "initialization or termination hypothesis."
    )
    distance_threshold = magicgui.widgets.create_widget(
        value=20.0,
        name="theta_dist",
        label="distance threshold",
        widget_type="FloatSpinBox",
        options={"tooltip": distance_threshold_tooltip},
    )

    time_threshold_tooltip = (
        "A threshold time from the beginning or end of movie to add "
        "an initialization or termination hypothesis."
    )
    time_threshold = magicgui.widgets.create_widget(
        value=5.0,
        name="theta_time",
        label="time threshold",
        widget_type="FloatSpinBox",
        options={"tooltip": time_threshold_tooltip},
    )

    apoptosis_threshold_tooltip = (
        "Number of apoptotic detections to be considered a genuine event.\n"
        "Detections are counted consecutively from the back of the track"
    )
    apoptosis_threshold = magicgui.widgets.create_widget(
        value=5,
        name="apop_thresh",
        label="apoptosis threshold",
        widget_type="SpinBox",
        options={"tooltip": apoptosis_threshold_tooltip},
    )

    return [
        distance_threshold,
        time_threshold,
        apoptosis_threshold,
    ]


def _create_bin_size_widgets() -> list[Widget]:
    """Create widget for setting bin sizes for the HypothesisModel"""

    distance_bin_size_tooltip = (
        "Isotropic spatial bin size for considering hypotheses.\n"
        "Larger bin sizes generate more hypothesese for each tracklet."
    )
    distance_bin_size = magicgui.widgets.create_widget(
        value=40.0,
        name="dist_thresh",
        label="distance bin size",
        widget_type="FloatSpinBox",
        options={"tooltip": distance_bin_size_tooltip},
    )

    time_bin_size_tooltip = (
        "Temporal bin size for considering hypotheses.\n"
        "Larger bin sizes generate more hypothesese for each tracklet."
    )
    time_bin_size = magicgui.widgets.create_widget(
        value=2.0,
        name="time_thresh",
        label="time bin size",
        widget_type="FloatSpinBox",
        options={"tooltip": time_bin_size_tooltip},
    )

    return [
        distance_bin_size,
        time_bin_size,
    ]


def create_hypothesis_model_widgets() -> list[Widget]:
    """Create widgets for setting parameters of the MotionModel"""

    hypothesis_model_label = magicgui.widgets.create_widget(
        label="<b>Hypothesis model</b>",  # bold label
        widget_type="Label",
        gui_only=True,
    )

    hypotheses_widgets = _create_hypotheses_widgets()
    scaling_factor_widgets = _create_scaling_factor_widgets()
    threshold_widgets = _create_threshold_widgets()
    bin_size_widgets = _create_bin_size_widgets()

    segmentation_miss_rate_tooltip = (
        "Miss rate for the segmentation.\n"
        "e.g. 1/100 segmentations incorrect gives a segmentation miss rate of 0.01."
    )
    segmentation_miss_rate = magicgui.widgets.create_widget(
        value=0.1,
        name="segmentation_miss_rate",
        label="miss rate",
        widget_type="FloatSpinBox",
        options={"tooltip": segmentation_miss_rate_tooltip},
    )

    relax_tooltip = (
        "Disable the time and distance thresholds.\n"
        "This means that tracks can initialize or terminate anywhere and"
        "at any time in the dataset."
    )
    relax = magicgui.widgets.create_widget(
        value=True,
        name="relax",
        label="relax thresholds",
        widget_type="CheckBox",
        options={"tooltip": relax_tooltip},
    )

    return [
        hypothesis_model_label,
        *hypotheses_widgets,
        *scaling_factor_widgets,
        *threshold_widgets,
        *bin_size_widgets,
        segmentation_miss_rate,
        relax,
    ]



from typing import List, Optional

import numpy as np
from pydantic import BaseModel, root_validator, validator

from . import constants
from .optimise.hypothesis import H_TYPES, PyHypothesisParams

__all__ = ["MotionModel", "ObjectModel", "HypothesisModel"]


def _check_symmetric(
    x: np.ndarray, rtol: float = 1e-5, atol: float = 1e-8
) -> bool:
    """Check that a matrix is symmetric by comparing with it's own transpose."""
    return np.allclose(x, x.T, rtol=rtol, atol=atol)


class MotionModel(BaseModel):
    r"""The `btrack` motion model.

    Parameters
    ----------
    name : str
        A name identifier for the model.
    measurements : int
        The number of measurements of the system (e.g. 3 for x, y, z).
    states : int
        The number of states of the system (typically >= measurements). The
        standard states for a constant velocity model are (x, y, z, dx, dy, dz),
        i.e. 6 in total for 3 measurements.
    A : array (states, states)
        State transition matrix.
    H : array (measurements, states)
        Observation matrix.
    P : array (states, states)
        Initial covariance estimate.
    G : array (1, states), optional
        Simplified estimated error in process. Is used to calculate Q using
        Q = G.T @ G. Either G or Q must be defined.
    Q : array (states, states), optional
        Full estimated error in process. Either G or Q must be defined.
    R : array (measurements, measurements)
        Estimated error in measurements.
    dt : float
        Time difference (always 1).
    accuracy : float
        Integration limits for calculating the probabilities.
    max_lost : int
        Number of frames without observation before marking as lost.
    prob_not_assign : float
        The default probability to not assign a track.

    Notes
    -----
    Uses a Kalman filter [1]_:

    'Is an algorithm which uses a series of measurements observed over time,
    containing noise (random variations) and other inaccuracies, and produces
    estimates of unknown variables that tend to be more precise than those that
    would be based on a single measurement alone.'

    Predicted estimate of state:

    .. math:: \hat{x}_{t\vert~t-1} = A_t \hat{x}_{t-1\vert~t-1}

    Predicted estimate of covariance:

    .. math:: P_{t\vert~t-1} = A_t P_{t-1\vert~t-1} A_t^{\top} + Q_t

    This is just a wrapper for the data with a few convenience functions
    thrown in. Matrices must be stored Fortran style, because Eigen uses
    column major and Numpy uses row major storage.

    References
    ----------
    .. [1] 'A new approach to linear filtering and prediction problems.'
      Kalman RE, 1960 Journal of Basic Engineering
    """

    measurements: int
    states: int
    A: np.ndarray
    H: np.ndarray
    P: np.ndarray
    R: np.ndarray
    G: Optional[np.ndarray] = None
    Q: Optional[np.ndarray] = None
    dt: float = 1.0
    accuracy: float = 2.0
    max_lost: int = constants.MAX_LOST
    prob_not_assign: float = constants.PROB_NOT_ASSIGN
    name: str = "Default"

    @validator("A", "H", "P", "R", "G", "Q", pre=True)
    def parse_arrays(cls, v):
        if isinstance(v, dict):
            m = v.get("matrix", None)
            s = v.get("sigma", 1.0)
            return np.asarray(m, dtype=float) * s
        return np.asarray(v, dtype=float)

    @validator("A")
    def reshape_A(cls, a, values):
        shape = (values["states"], values["states"])
        return np.reshape(a, shape)

    @validator("H")
    def reshape_H(cls, h, values):
        shape = (values["measurements"], values["states"])
        return np.reshape(h, shape)

    @validator("P")
    def reshape_P(cls, p, values):
        shape = (values["states"], values["states"])
        p = np.reshape(p, shape)
        if not _check_symmetric(p):
            raise ValueError("Matrix `P` is not symmetric.")
        return p

    @validator("R")
    def reshape_R(cls, r, values):
        shape = (values["measurements"], values["measurements"])
        r = np.reshape(r, shape)
        if not _check_symmetric(r):
            raise ValueError("Matrix `R` is not symmetric.")
        return r

    @validator("G")
    def reshape_G(cls, g, values):
        shape = (1, values["states"])
        return np.reshape(g, shape)

    @validator("Q")
    def reshape_Q(cls, q, values):
        shape = (values["states"], values["states"])
        q = np.reshape(q, shape)
        if not _check_symmetric(q):
            raise ValueError("Matrix `Q` is not symmetric.")
        return q

    @root_validator
    def validate_motion_model(cls, values):
        if values["Q"] is None:
            G = values.get("G", None)
            if G is None:
                raise ValueError("Either a `G` or `Q` matrix is required.")
            values["Q"] = G.T @ G
        return values

    class Config:
        arbitrary_types_allowed = True
        validate_assignment = True


class ObjectModel(BaseModel):
    """The `btrack` object model.

    This is a class to deal with state transitions in the object, essentially
    a Hidden Markov Model.  Makes an assumption that the states are all
    observable, but with noise.

    Parameters
    ----------
    name : str
        A name identifier for the model.
    emission : array
        The emission probability matrix.
    transition : array
        Transition probabilities.
    start : array
        Initial probabilities.
    states : int
        Number of observable states.
    """

    states: int
    emission: np.ndarray
    transition: np.ndarray
    start: np.ndarray
    name: str = "Default"

    @validator("emission", "transition", "start", pre=True)
    def parse_array(cls, v, values):
        return np.asarray(v, dtype=float)

    @validator("emission", "transition", "start", pre=True)
    def reshape_emission_transition(cls, v, values):
        shape = (values["states"], values["states"])
        return np.reshape(v, shape)

    @validator("emission", "transition", "start", pre=True)
    def reshape_start(cls, v, values):
        shape = (1, values["states"])
        return np.reshape(v, shape)

    class Config:
        arbitrary_types_allowed = True
        validate_assignment = True


class HypothesisModel(BaseModel):
    r"""The `btrack` hypothesis model.

    This is a class to deal with hypothesis generation in the optimization step
    of the tracking algorithm.

    Parameters
    ----------
    name : str
        A name identifier for the model.
    hypotheses : list[str]
        A list of hypotheses to be generated. See `optimise.hypothesis.H_TYPES`.
    lambda_time : float
        A scaling factor for the influence of time when determining
        initialization or termination hypotheses. See notes.
    lambda_dist : float
        A a scaling factor for the influence of distance at the border when
        determining initialization or termination hypotheses. See notes.
    lambda_link : float
        A scaling factor for the influence of track-to-track distance on linking
        probability. See notes.
    lambda_branch : float
        A scaling factor for the influence of cell state and position on
        division (mitosis/branching) probability. See notes.
    eta : float
        Default value for a low probability event (e.g. 1E-10) to prevent
        divide-by-zero.
    theta_dist : float
        A threshold distance from the edge of the FOV to add an initialization
        or termination hypothesis.
    theta_time : float
        A threshold time from the beginning or end of movie to add an
        initialization or termination hypothesis.
    dist_thresh : float
        Isotropic spatial bin size for considering hypotheses. Larger bin sizes
        generate more hypothesese for each tracklet.
    time_thresh : float
        Temporal bin size for considering hypotheses. Larger bin sizes generate
        more hypothesese for each tracklet.
    apop_thresh : int
        Number of apoptotic detections, counted consecutively from the back of
        the track, to be considered a real apoptosis.
    segmentation_miss_rate : float
        Miss rate for the segmentation, e.g. 1/100 segmentations incorrect gives
        a segmentation miss rate or 0.01.
    apoptosis_rate : float
        Rate of apoptosis detections.
    relax : bool
        Disables the `theta_dist` and `theta_time` thresholds when creating
        termination and initialization hypotheses. This means that tracks can
        initialize or terminate anywhere (or time) in the dataset.

    Notes
    -----
    The `lambda` (:math:`\lambda`) factors scale the probability according to
    the following function:

    .. math:: e^{(-d / \lambda)}
    """

    hypotheses: List[str]
    lambda_time: float
    lambda_dist: float
    lambda_link: float
    lambda_branch: float
    eta: float
    theta_dist: float
    theta_time: float
    dist_thresh: float
    time_thresh: float
    apop_thresh: int
    segmentation_miss_rate: float
    apoptosis_rate: float
    relax: bool
    name: str = "Default"

    @validator("hypotheses", pre=True)
    def parse_hypotheses(cls, hypotheses):
        if not all(h in H_TYPES for h in hypotheses):
            raise ValueError("Unknown hypothesis type in `hypotheses`.")
        return hypotheses

    def hypotheses_to_generate(self) -> int:
        """Return an integer representation of the hypotheses to generate."""
        h_bin = "".join(
            [str(int(h)) for h in [h in self.hypotheses for h in H_TYPES]]
        )
        return int(h_bin[::-1], 2)

    def as_ctype(self) -> PyHypothesisParams:
        """Return the ctypes representation of the `HypothesisModel`."""
        h_params = PyHypothesisParams()
        fields = [f[0] for f in h_params._fields_]

        for k, v in self.dict().items():
            if k in fields:
                setattr(h_params, k, v)

        # set the hypotheses to generate
        h_params.hypotheses_to_generate = self.hypotheses_to_generate()
        return h_params

    class Config:
        validate_assignment = True



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from magicgui.widgets import Widget

import magicgui


def _make_label_bold(label: str) -> str:
    """Generate html for a bold label"""

    return f"<b>{label}</b>"


def _create_sigma_widgets() -> list[Widget]:
    """Create widgets for setting the magnitudes of the MotionModel matrices"""

    P_sigma_tooltip = (
        "Magnitude of error in initial estimates.\n"
        "Used to scale the matrix P."
    )
    P_sigma = magicgui.widgets.create_widget(
        value=150.0,
        name="P_sigma",
        label=f"max({_make_label_bold('P')})",
        widget_type="FloatSpinBox",
        options={"tooltip": P_sigma_tooltip},
    )

    G_sigma_tooltip = (
        "Magnitude of error in process.\n Used to scale the matrix G."
    )
    G_sigma = magicgui.widgets.create_widget(
        value=15.0,
        name="G_sigma",
        label=f"max({_make_label_bold('G')})",
        widget_type="FloatSpinBox",
        options={"tooltip": G_sigma_tooltip},
    )

    R_sigma_tooltip = (
        "Magnitude of error in measurements.\n Used to scale the matrix R."
    )
    R_sigma = magicgui.widgets.create_widget(
        value=5.0,
        name="R_sigma",
        label=f"max({_make_label_bold('R')})",
        widget_type="FloatSpinBox",
        options={"tooltip": R_sigma_tooltip},
    )

    return [
        P_sigma,
        G_sigma,
        R_sigma,
    ]


def create_motion_model_widgets() -> list[Widget]:
    """Create widgets for setting parameters of the MotionModel"""

    motion_model_label = magicgui.widgets.create_widget(
        label=_make_label_bold("Motion model"),
        widget_type="Label",
        gui_only=True,
    )

    sigma_widgets = _create_sigma_widgets()

    accuracy_tooltip = "Integration limits for calculating probabilities"
    accuracy = magicgui.widgets.create_widget(
        value=7.5,
        name="accuracy",
        label="accuracy",
        widget_type="FloatSpinBox",
        options={"tooltip": accuracy_tooltip},
    )

    max_lost_frames_tooltip = (
        "Number of frames without observation before marking as lost"
    )
    max_lost_frames = magicgui.widgets.create_widget(
        value=5,
        name="max_lost",
        label="max lost",
        widget_type="SpinBox",
        options={"tooltip": max_lost_frames_tooltip},
    )

    return [
        motion_model_label,
        *sigma_widgets,
        accuracy,
        max_lost_frames,
    ]



from __future__ import annotations

import logging
from typing import Any, Dict, List, Union

import numpy as np

# import core
from btrack import btypes, constants

# get the logger instance
logger = logging.getLogger(__name__)


def localizations_to_objects(
    localizations: Union[
        np.ndarray, List[btypes.PyTrackObject], Dict[str, Any]
    ]
) -> List[btypes.PyTrackObject]:
    """Take a numpy array or pandas dataframe and convert to PyTrackObjects.

    Parameters
    ----------
    localizations : list[PyTrackObject], np.ndarray, pandas.DataFrame
        A list or array of localizations.

    Returns
    -------
    objects : list[PyTrackObject]
        A list of PyTrackObject objects that represent the localizations.
    """

    logger.info(f"Objects are of type: {type(localizations)}")

    if isinstance(localizations, list) and check_object_type(localizations):
        # if these are already PyTrackObjects just silently return
        return localizations

    # do we have a numpy array or pandas dataframe?
    if isinstance(localizations, np.ndarray):
        return objects_from_array(localizations)
    else:
        try:
            objects_dict = {
                c: np.asarray(localizations[c]) for c in localizations
            }
        except ValueError as err:
            logger.error(f"Unknown localization type: {type(localizations)}")
            raise TypeError(
                f"Unknown localization type: {type(localizations)}"
            ) from err

    # how many objects are there
    n_objects = objects_dict["t"].shape[0]
    objects_dict["ID"] = np.arange(n_objects)

    return objects_from_dict(objects_dict)


def objects_from_dict(objects_dict: dict) -> List[btypes.PyTrackObject]:
    """Construct PyTrackObjects from a dictionary"""
    # now that we have the object dictionary, convert this to objects
    objects = []
    n_objects = int(objects_dict["t"].shape[0])

    assert all([v.shape[0] == n_objects for k, v in objects_dict.items()])

    for i in range(n_objects):
        data = {k: v[i] for k, v in objects_dict.items()}
        obj = btypes.PyTrackObject.from_dict(data)
        objects.append(obj)
    return objects


def objects_from_array(
    objects_arr: np.ndarray,
    *,
    default_keys: List[str] = constants.DEFAULT_OBJECT_KEYS,
) -> List[btypes.PyTrackObject]:
    """Construct PyTrackObjects from a numpy array."""
    assert objects_arr.ndim == constants.Dimensionality.TWO

    MIN_N_FEATURES = 3
    n_features = objects_arr.shape[1]
    assert n_features >= MIN_N_FEATURES

    n_objects = objects_arr.shape[0]

    keys = default_keys[:n_features]
    objects_dict = {keys[i]: objects_arr[:, i] for i in range(n_features)}
    objects_dict["ID"] = np.arange(n_objects)
    return objects_from_dict(objects_dict)


def check_track_type(tracks: list) -> bool:
    return all(isinstance(t, btypes.Tracklet) for t in tracks)


def check_object_type(objects: list) -> bool:
    return all(isinstance(o, btypes.PyTrackObject) for o in objects)



from __future__ import annotations

import csv
import logging
import os
from typing import TYPE_CHECKING, Optional

import numpy as np

from btrack import constants

from .hdf import HDF5FileHandler
from .utils import check_track_type

if TYPE_CHECKING:
    from btrack import BayesianTracker

# get the logger instance
logger = logging.getLogger(__name__)


def export_delegator(
    filename: os.PathLike,
    tracker: BayesianTracker,
    obj_type: Optional[str] = None,
    filter_by: Optional[str] = None,
) -> None:
    """Export data from the tracker using the appropriate exporter.

    Parameters
    ----------
    filename : str
        The filename to export the data. The extension (e.g. .h5) is used
        to select the correct export function.
    tracker : BayesianTracker
        An instance of the tracker.
    obj_type : str, optional
        The object type to export the data. Usually `obj_type_1`
    filter_by : str, optional
        A string that represents how the data has been filtered prior to
        tracking, e.g. using the object property `area>100`

    Notes
    -----
    This uses the appropriate exporter dependent on the given file extension.
    """
    export_dir, export_fn = os.path.split(filename)
    _, ext = os.path.splitext(filename)

    if ext == ".csv":
        export_CSV(filename, tracker.tracks, obj_type=obj_type)
    elif ext in (".hdf", ".hdf5", ".h5"):
        _export_HDF(filename, tracker, obj_type=obj_type, filter_by=filter_by)
    else:
        logger.error(f"Export file format {ext} not recognized.")


def export_CSV(
    filename: os.PathLike,
    tracks: list,
    properties: list = constants.DEFAULT_EXPORT_PROPERTIES,
    obj_type: Optional[str] = None,
):
    """Export the track data as a simple CSV file.

    Parameters
    ----------
    filename : str
        The filename of the file to be exported.
    tracks : list[Tracklet]
        A list of Tracklet objects to be exported.
    properties : list, default = constants.DEFAULT_EXPORT_PROPERTIES
        A list of tracklet properties to be exported.
    obj_type : str, optional
        A string describing the object type, e.g. `obj_type_1`.

    """

    if not tracks:
        logger.error(f"No tracks found when exporting to: {filename}")
        return

    if not check_track_type(tracks):
        logger.error("Tracks of incorrect type")

    logger.info(f"Writing out CSV files to: {filename}")
    export_track = np.vstack([t.to_array(properties) for t in tracks])

    with open(filename, "w", newline="") as csv_file:
        csvwriter = csv.writer(csv_file, delimiter=" ")
        csvwriter.writerow(properties)
        for i in range(export_track.shape[0]):
            csvwriter.writerow(export_track[i, :].tolist())


def export_LBEP(filename: os.PathLike, tracks: list):
    """Export the LBEP table as a text file."""
    if not tracks:
        logger.error(f"No tracks found when exporting to: {filename}")
        return

    if not check_track_type(tracks):
        logger.error("Tracks of incorrect type")

    tracks.sort(key=lambda t: t.ID)

    with open(filename, "w") as lbep_file:
        logger.info(f"Writing LBEP file: {filename}...")
        for track in tracks:
            lbep = f"{track.ID} {track.t[0]} {track.t[-1]} {track.parent}"
            lbep_file.write(f"{lbep}\n")


def _export_HDF(
    filename: os.PathLike, tracker, obj_type=None, filter_by: str = None
):
    """Export to HDF."""

    filename_noext, ext = os.path.splitext(filename)
    if ext != ".h5":
        filename = filename_noext + ".h5"
        logger.warning(f"Changing HDF filename to {filename}")

    with HDF5FileHandler(filename, read_write="a", obj_type=obj_type) as hdf:
        # if there are no objects, write them out
        if f"objects/{obj_type}" not in hdf._hdf:
            hdf.write_objects(tracker)
        # write the tracks
        hdf.write_tracks(tracker, f_expr=filter_by)



import json

import numpy as np
import pytest

from ._utils import (
    TEST_DATA_PATH,
    full_tracker_example,
    simple_tracker_example,
)


def _gt_object_hook(d):
    """JSON stores keys as strings, convert these to integers."""
    return {int(k): v for k, v in d.items()}


def _load_ground_truth():
    with open(TEST_DATA_PATH / "test_ground_truth.json", "r") as file:
        ground_truth = json.load(file, object_hook=_gt_object_hook)
    return ground_truth


def _load_ground_truth_graph() -> dict:
    with open(TEST_DATA_PATH / "test_graph.json", "r") as file:
        ground_truth_graph = json.load(file, object_hook=_gt_object_hook)

    return ground_truth_graph


def _get_tracklet(tracks: dict, idx: int) -> list:
    """Get a tracklet by the first object ID"""
    target = [t for t in tracks.values() if t[0] == idx]
    if target:
        return target[0]
    else:
        raise ValueError("Object ID not found.")


def test_tracker(test_real_objects):
    """Test the tracks output of the tracker, using the default config and known
    data."""
    ground_truth = _load_ground_truth()

    tracker = full_tracker_example(test_real_objects)
    tracks = tracker.tracks

    # iterate over the tracks and check that the object references match
    for track in tracks:
        gt_refs = _get_tracklet(ground_truth, track.refs[0])
        np.testing.assert_equal(track.refs, gt_refs)


def test_tracker_kalman(test_real_objects):
    """Test the Kalman filter output of the tracker."""
    tracker = full_tracker_example(
        test_real_objects,
    )
    tracker.return_kalman = True
    assert tracker.return_kalman

    tracks = tracker.tracks

    for track in tracks:
        assert track.kalman.shape[0] == len(track)
        assert track.mu(0).shape == (3, 1)
        assert track.covar(0).shape == (3, 3)
        assert track.predicted(0).shape == (3, 1)

        np.testing.assert_equal(
            np.array([track.x[0], track.y[0], track.z[0]]),
            np.squeeze(track.mu(0)),
        )


def test_tracker_graph(test_real_objects):
    """Test the graph output of the tracker, using the default config and known
    data."""

    ground_truth_graph = _load_ground_truth_graph()

    # run the tracking
    tracker = full_tracker_example(test_real_objects)
    _, _, graph = tracker.to_napari(ndim=2)

    assert ground_truth_graph == graph


def test_tracker_frames():
    """Test to make sure all frames are accounted for."""

    tracker, objects = simple_tracker_example()
    tracks = tracker.tracks

    assert len(tracks) == 1
    track = tracks[0]
    np.testing.assert_equal(track.t, objects["t"])


@pytest.mark.parametrize("store_graph", [False, True])
def test_tracker_candidate_graph(test_real_objects, store_graph):
    """Check that a candidate graph is returned if we set the flag to store it."""
    tracker = full_tracker_example(
        test_real_objects,
        store_candidate_graph=store_graph,
    )
    assert tracker.store_candidate_graph == store_graph
    edges = tracker.candidate_graph_edges()

    # graph should contain edges
    assert (
        bool(edges) == store_graph
    ), f"Found {len(edges)} edges in candidate graph."



import warnings

warnings.warn(  # noqa: B028
    "`btrack.dataio` has been deprecated. Please use `btrack.io` subpackage "
    "instead.",
    # DeprecationWarning,
)

from .io import HDF5FileHandler  # noqa: F401,E402



import os
from typing import List

import numpy as np
import pooch
from skimage.io import imread

from .btypes import PyTrackObject
from .io import import_CSV

BASE_URL = (
    "https://raw.githubusercontent.com/lowe-lab-ucl/btrack-examples/main/"
)

CACHE_PATH = pooch.os_cache("btrack-examples")


def _remote_registry() -> os.PathLike:
    file_path = pooch.retrieve(
        # URL to one of Pooch's test files
        path=CACHE_PATH,
        url=BASE_URL + "registry.txt",
        known_hash="673de62c62eeb6f356fb1bff968748566d23936f567201cf61493d031d42d480",
    )
    return file_path


POOCH = pooch.create(
    path=CACHE_PATH,
    base_url=BASE_URL,
    version_dev="main",
    registry=None,
)
POOCH.load_registry(_remote_registry())


def cell_config() -> os.PathLike:
    """Return the file path to the example `cell_config`."""
    file_path = POOCH.fetch("examples/cell_config.json")
    return file_path


def particle_config() -> os.PathLike:
    """Return the file path to the example `particle_config`."""
    file_path = POOCH.fetch("examples/particle_config.json")
    return file_path


def example_segmentation_file() -> os.PathLike:
    """Return the file path to the example U-Net segmentation image file."""
    file_path = POOCH.fetch("examples/segmented.tif")
    return file_path


def example_segmentation() -> np.ndarray:
    """Return the U-Net segmentation as a numpy array of dimensions (T, Y, X)."""
    file_path = example_segmentation_file()
    segmentation = imread(file_path)
    return segmentation


def example_track_objects_file() -> os.PathLike:
    """Return the file path to the example localized and classified objects
    stored in a CSV file."""
    file_path = POOCH.fetch("examples/objects.csv")
    return file_path


def example_track_objects() -> List[PyTrackObject]:
    """Return the example localized and classified objects stored in a CSV file
    as a list `PyTrackObject`s to be used by the tracker."""
    file_path = example_track_objects_file()
    objects = import_CSV(file_path)
    return objects



from __future__ import annotations

import inspect
import logging
from typing import Generator, List, Optional, Tuple, Union

import numpy as np
from skimage.measure import label, regionprops_table

from btrack import btypes
from btrack.constants import Dimensionality

from .utils import localizations_to_objects

# get the logger instance
logger = logging.getLogger(__name__)


def _centroids_from_single_arr(
    segmentation: Union[np.ndarray, Generator],
    properties: Tuple[str],
    frame: int,
    intensity_image: Optional[np.ndarray] = None,
    scale: Optional[Tuple[float]] = None,
    *,
    use_weighted_centroid: bool = False,
    assign_class_ID: bool = False,
) -> np.ndarray:
    """Return the object centroids from a numpy array representing the
    image data."""

    if np.sum(segmentation) == 0:
        return {}

    def _is_unique(x: np.ndarray) -> bool:
        # check if image is uniquely labelled (necessary for regionprops)
        return np.max(label(x)) == np.max(x)

    if use_weighted_centroid and intensity_image is not None:
        CENTROID_PROPERTY = "weighted_centroid"
    else:
        CENTROID_PROPERTY = "centroid"

    if CENTROID_PROPERTY not in properties:
        properties = (CENTROID_PROPERTY, *properties)

    # if class id is specified then extract that property first
    if assign_class_ID:
        # ensure regionprops can properly read label image
        labeled = label(segmentation)

        # pull class_ID from segments using pixel intensity
        _class_ID_centroids = regionprops_table(
            labeled,
            intensity_image=segmentation,
            properties=("max_intensity",),
        )

        # rename class_ID column and remove keyword from properties
        _class_ID_centroids["class_id"] = _class_ID_centroids.pop(
            "max_intensity"
        )

        # run regionprops to record other intensity image properties
        _centroids = regionprops_table(
            labeled,
            intensity_image=intensity_image,
            properties=properties,
        )

        # merge centroids with class ID centroids
        _centroids.update(_class_ID_centroids)

        assert "class_id" in _centroids, _centroids.keys()

    else:
        # check to see whether the segmentation is unique
        labeled = (
            label(segmentation)
            if not _is_unique(segmentation)
            else segmentation
        )

        _centroids = regionprops_table(
            labeled,
            intensity_image=intensity_image,
            properties=properties,
        )

    # add time to the array
    _centroids["t"] = np.full(
        _centroids[f"{CENTROID_PROPERTY}-0"].shape, frame
    )

    # apply the anistropic scaling
    if scale is not None:
        if len(scale) != segmentation.ndim:
            raise ValueError("Scale dimensions do not match segmentation.")

        # perform the anistropic scaling
        for dim in range(segmentation.ndim):
            _centroids[f"{CENTROID_PROPERTY}-{dim}"] = np.multiply(
                _centroids[f"{CENTROID_PROPERTY}-{dim}"], float(scale[dim])
            )

    # now rename the axes for btrack
    dim_names = ["z", "y", "x"][-(segmentation.ndim) :]
    for dim in range(segmentation.ndim):
        dim_name = dim_names[dim]
        _centroids[dim_name] = _centroids.pop(f"{CENTROID_PROPERTY}-{dim}")

    return _centroids


def _concat_centroids(centroids, new_centroids):
    """Concatentate centroid dictionaries."""
    for key, values in new_centroids.items():
        if key not in centroids:
            centroids[key] = values
        else:
            centroids[key] = np.concatenate([centroids[key], values])
    return centroids


def segmentation_to_objects(  # noqa: PLR0913
    segmentation: Union[np.ndarray, Generator],
    intensity_image: Optional[Union[np.ndarray, Generator]] = None,
    properties: Optional[Tuple[str]] = (),
    scale: Optional[Tuple[float]] = None,
    use_weighted_centroid: bool = True,  # noqa: FBT001,FBT002
    assign_class_ID: bool = False,  # noqa: FBT001,FBT002
) -> List[btypes.PyTrackObject]:
    """Convert segmentation to a set of trackable objects.

    Parameters
    ----------
    segmentation : np.ndarray, dask.array.core.Array or Generator
        Segmentation can be provided in several different formats. Arrays should
        be ordered as T(Z)YX.
    intensity_image : np.ndarray, dask.array.core.Array or Generator, optional
        Intensity image with same size as segmentation, to be used to calculate
        additional properties. See `skimage.measure.regionprops` for more info.
    properties : tuple of str, optional
        Properties passed to scikit-image regionprops. These additional
        properties are added as metadata to the btrack objects.
        See `skimage.measure.regionprops` for more info.
    scale : tuple
        A scale for each spatial dimension of the input segmentation. Defaults
        to one for all axes, and allows scaling for anisotropic imaging data.
    use_weighted_centroid : bool, default True
        If an intensity image has been provided, default to calculating the
        weighted centroid. See `skimage.measure.regionprops` for more info.
    assign_class_ID : bool, default False
        If specified, assign a class label for each individual object based on
        the pixel intensity found in the mask. Requires semantic segmentation,
        i.e. object type 1 will have pixel value 1.

    Returns
    -------
    objects : list
        A list of :py:meth:`btrack.btypes.PyTrackObject` trackable objects.

    Examples
    --------
    >>> objects = btrack.utils.segmentation_to_objects(
    ...   segmentation,
    ...   properties=('area', ),
    ...   scale=(1., 1.),
    ...   assign_class_ID=True,
    ... )
    """

    centroids: dict = {}
    USE_INTENSITY = False
    USE_WEIGHTED = False

    logger.info("Localizing objects from segmentation...")

    # if we have an intensity image, add that here
    if intensity_image is not None:
        if type(intensity_image) != type(segmentation):
            raise TypeError(
                "Segmentation and intensity image must be the same type."
            )
        USE_INTENSITY = True
        USE_WEIGHTED = use_weighted_centroid and USE_INTENSITY

    if USE_INTENSITY:
        logger.info("Found intensity_image data")

    if USE_WEIGHTED:
        logger.info("Calculating weighted centroids using intensity_image")

    # we need to remove 'label' since this is a protected keyword for btrack
    # objects
    if isinstance(properties, tuple) and "label" in properties:
        logger.warning("Cannot use scikit-image label as a property.")
        properties = set(properties)
        properties.remove("label")
        properties = tuple(properties)

    if inspect.isgeneratorfunction(segmentation) or isinstance(
        segmentation, Generator
    ):
        for frame, seg in enumerate(segmentation):
            intens = next(intensity_image) if USE_INTENSITY else None
            _centroids = _centroids_from_single_arr(
                seg,
                properties,
                frame,
                intensity_image=intens,
                scale=scale,
                use_weighted_centroid=USE_WEIGHTED,
                assign_class_ID=assign_class_ID,
            )

            # concatenate the centroids
            centroids = _concat_centroids(centroids, _centroids)

    else:
        if segmentation.ndim not in (
            Dimensionality.THREE,
            Dimensionality.FOUR,
        ):
            raise ValueError("Segmentation array must have 3 or 4 dims.")

        for frame in range(segmentation.shape[0]):
            # try to cast to numpy array, should work for dask arrays and implicitly
            # call the `.compute()` method
            seg = np.asarray(segmentation[frame, ...])
            intens = (
                np.asarray(intensity_image[frame, ...])
                if USE_INTENSITY
                else None
            )
            _centroids = _centroids_from_single_arr(
                seg,
                properties,
                frame,
                intensity_image=intens,
                scale=scale,
                use_weighted_centroid=USE_WEIGHTED,
                assign_class_ID=assign_class_ID,
            )

            # concatenate the centroids
            centroids = _concat_centroids(centroids, _centroids)

    if not centroids:
        logger.warning("...Found no objects.")
        return []

    # now create the btrack objects
    objects = localizations_to_objects(centroids)
    n_frames = int(np.max(centroids["t"]) + 1)

    logger.info(f"...Found {len(objects)} objects in {n_frames} frames.")

    return objects



"""
Show the btrack widget with example data
=========================
This example:
- loads a sample segmentation and cell config
- adds the segmentation to a napari viewer
- loads the btrack plugin
- opens the napari viewer
"""
import napari

from btrack import datasets

viewer = napari.Viewer()
napari.current_viewer()

_, btrack_widget = viewer.window.add_plugin_dock_widget(
    plugin_name="napari-btrack", widget_name="Track"
)


segmentation = datasets.example_segmentation()
viewer.add_labels(segmentation)
# napari takes the first image layer as default anyway here, but better to be explicit
btrack_widget.segmentation.choices = viewer.layers
btrack_widget.segmentation.value = viewer.layers["segmentation"]

if __name__ == "__main__":
    # The napari event loop needs to be run under here to allow the window
    # to be spawned from a Python script
    napari.run()



"""
This module contains functions for syncing widget values with TrackerConfig
values.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from magicgui.widgets import Container

    from btrack.config import TrackerConfig
    from btrack.napari.config import Sigmas, UnscaledTrackerConfig

import btrack.napari.constants


def update_config_from_widgets(
    unscaled_config: UnscaledTrackerConfig,
    container: Container,
) -> TrackerConfig:
    """Update an UnscaledTrackerConfig with the current widget values."""

    # Update MotionModel matrix scaling factors
    sigmas: Sigmas = unscaled_config.sigmas
    for matrix_name in sigmas:
        sigmas[matrix_name] = container[f"{matrix_name}_sigma"].value

    # Update TrackerConfig values
    config = unscaled_config.tracker_config
    update_method_name = container.update_method.current_choice
    update_method_index = container.update_method.choices.index(
        update_method_name
    )
    config.update_method = update_method_index
    config.max_search_radius = container.max_search_radius.value

    # Update MotionModel values
    motion_model = config.motion_model
    motion_model.accuracy = container.accuracy.value
    motion_model.max_lost = container.max_lost.value

    # Update HypothesisModel.hypotheses values
    hypothesis_model = config.hypothesis_model
    hypothesis_model.hypotheses = [
        hypothesis
        for hypothesis in btrack.napari.constants.HYPOTHESES
        if container[hypothesis].value
    ]

    # Update HypothesisModel scaling factors
    for scaling_factor in btrack.napari.constants.HYPOTHESIS_SCALING_FACTORS:
        setattr(
            hypothesis_model, scaling_factor, container[scaling_factor].value
        )

    # Update HypothesisModel thresholds
    for threshold in btrack.napari.constants.HYPOTHESIS_THRESHOLDS:
        setattr(hypothesis_model, threshold, container[threshold].value)

    hypothesis_model.segmentation_miss_rate = (
        container.segmentation_miss_rate.value
    )

    return unscaled_config


def update_widgets_from_config(
    unscaled_config: UnscaledTrackerConfig,
    container: Container,
) -> Container:
    """
    Update the widgets in a container with the values in an
    UnscaledTrackerConfig.
    """

    # Update widgets from MotionModel matrix scaling factors
    sigmas: Sigmas = unscaled_config.sigmas
    for matrix_name in sigmas:
        container[f"{matrix_name}_sigma"].value = sigmas[matrix_name]

    # Update widgets from TrackerConfig values
    config = unscaled_config.tracker_config
    container.update_method.value = config.update_method.name
    container.max_search_radius.value = config.max_search_radius

    # Update widgets from MotionModel values
    motion_model = config.motion_model
    container.accuracy.value = motion_model.accuracy
    container.max_lost.value = motion_model.max_lost

    # Update widgets from HypothesisModel.hypotheses values
    hypothesis_model = config.hypothesis_model
    for hypothesis in btrack.napari.constants.HYPOTHESES:
        is_checked = hypothesis in hypothesis_model.hypotheses
        container[hypothesis].value = is_checked

    # Update widgets from HypothesisModel scaling factors
    for scaling_factor in btrack.napari.constants.HYPOTHESIS_SCALING_FACTORS:
        container[scaling_factor].value = getattr(
            hypothesis_model, scaling_factor
        )

    # Update widgets from HypothesisModel thresholds
    for threshold in btrack.napari.constants.HYPOTHESIS_THRESHOLDS:
        container[threshold].value = getattr(hypothesis_model, threshold)

    container.segmentation_miss_rate.value = (
        hypothesis_model.segmentation_miss_rate
    )

    return container



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from magicgui.widgets import Widget

import magicgui
import napari


def create_input_widgets() -> list[Widget]:
    """Create widgets for selecting labels layer and TrackerConfig"""

    segmentation_tooltip = (
        "Select a 'Labels' layer to use for tracking.\n"
        "To use an 'Image' layer, first convert 'Labels' by right-clicking "
        "on it in the layers list, and clicking on 'Convert to Labels'"
    )
    segmentation = magicgui.widgets.create_widget(
        annotation=napari.layers.Labels,
        name="segmentation",
        label="segmentation",
        options={"tooltip": segmentation_tooltip},
    )

    config_tooltip = (
        "Select a loaded configuration.\n"
        "Note, this will update values set below."
    )
    config = magicgui.widgets.create_widget(
        value="cell",
        name="config",
        label="config name",
        widget_type="ComboBox",
        options={
            "choices": ["cell", "particle"],
            "tooltip": config_tooltip,
        },
    )

    return [segmentation, config]


def create_update_method_widgets() -> list[Widget]:
    """Create widgets for selecting the update method"""

    update_method_tooltip = (
        "Select the update method.\n"
        "EXACT: exact calculation of Bayesian belief matrix.\n"
        "APPROXIMATE: approximate the Bayesian belief matrix. Useful for datasets with "
        "more than 1000 particles per frame."
    )
    update_method = magicgui.widgets.create_widget(
        value="EXACT",
        name="update_method",
        label="update method",
        widget_type="ComboBox",
        options={
            "choices": ["EXACT", "APPROXIMATE"],
            "tooltip": update_method_tooltip,
        },
    )

    # TODO: this widget should be hidden when the update method is set to EXACT
    max_search_radius_tooltip = (
        "The local spatial search radius (isotropic, pixels) used when the update "
        "method is 'APPROXIMATE'"
    )
    max_search_radius = magicgui.widgets.create_widget(
        value=100,
        name="max_search_radius",
        label="search radius",
        widget_type="SpinBox",
        options={"tooltip": max_search_radius_tooltip},
    )

    return [update_method, max_search_radius]


def create_control_widgets() -> list[Widget]:
    """Create widgets for running the analysis or handling I/O.

    This includes widgets for running the tracking, saving and loading
    configuration files, and resetting the widget values to those in
    the selected config."""

    names = [
        "load_config_button",
        "save_config_button",
        "reset_button",
        "call_button",
    ]
    labels = [
        "Load configuration",
        "Save configuration",
        "Reset defaults",
        "Run",
    ]
    tooltips = [
        "Load a TrackerConfig json file.",
        "Export the current configuration to a TrackerConfig json file.",
        "Reset the current configuration to the defaults stored in the corresponding json file.",  # noqa: E501
        "Run the tracking analysis with the current configuration.",
    ]

    control_buttons = []
    for name, label, tooltip in zip(names, labels, tooltips):
        widget = magicgui.widgets.create_widget(
            name=name,
            label=label,
            widget_type="PushButton",
            options={"tooltip": tooltip},
        )
        control_buttons.append(widget)

    return control_buttons



from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
from skimage.measure import label

import btrack

CONFIG_FILE = (
    Path(__file__).resolve().parent.parent / "models" / "cell_config.json"
)

TEST_DATA_PATH = Path(__file__).resolve().parent / "_test_data"

RANDOM_SEED = 1234


def create_test_object(
    test_id: Optional[int] = None,
    ndim: int = 3,
) -> Tuple[btrack.btypes.PyTrackObject, Dict[str, Any]]:
    """Create a test object."""

    rng = np.random.default_rng(seed=RANDOM_SEED)

    data = {
        "ID": rng.integers(0, 1000) if test_id is None else int(test_id),
        "x": rng.uniform(0.0, 1000.0),
        "y": rng.uniform(0.0, 1000.0),
        "z": rng.uniform(0.0, 1000.0)
        if ndim == btrack.constants.Dimensionality.THREE
        else 0.0,
        "t": rng.integers(0, 1000),
        "dummy": False,
        "label": 0,
    }

    obj = btrack.btypes.PyTrackObject().from_dict(data)
    return obj, data


def create_test_properties() -> Dict[str, float]:
    """Create test properties for an object."""
    rng = np.random.default_rng(seed=RANDOM_SEED)
    properties = {
        "speed": rng.uniform(0.0, 1.0),
        "circularity": rng.uniform(0.0, 1.0),
        "reporter": rng.uniform(0.0, 1.0),
        "nD": rng.uniform(0.0, 1.0, size=(5,)),
    }
    return properties


def create_test_tracklet(
    track_len: int,
    track_id: Optional[int] = None,
    ndim: int = 3,
) -> Tuple[
    btrack.btypes.Tracklet,
    List[btrack.btypes.PyTrackObject],
    List[Dict[str, Any]],
    int,
]:
    """Create a test track."""
    rng = np.random.default_rng(seed=RANDOM_SEED)

    data = [create_test_object(ndim=ndim)[0] for _ in range(track_len)]
    props = [create_test_properties() for _ in range(track_len)]
    for idx, obj in enumerate(data):
        obj.properties = props[idx]
    track_id = rng.integers(0, 1000) if track_id is None else track_id
    tracklet = btrack.btypes.Tracklet(track_id, data)
    tracklet.parent = track_id
    tracklet.root = track_id

    # convert to dictionary {key: [p0,...,pn]}
    properties = (
        {} if not props else {k: [p[k] for p in props] for k in props[0]}
    )

    return tracklet, data, properties, track_id


def create_realistic_tracklet(  # noqa: PLR0913
    start_x: float,
    start_y: float,
    dx: float,
    dy: float,
    track_len: float,
    track_ID: int,
) -> btrack.btypes.Tracklet:
    """Create a realistic moving track."""

    data = {
        "x": np.array([start_x + dx * t for t in range(track_len)]),
        "y": np.array([start_y + dy * t for t in range(track_len)]),
        "t": np.arange(track_len),
        "ID": np.array(
            [(track_ID - 1) * track_len + t for t in range(track_len)]
        ),
    }

    objects = btrack.io.objects_from_dict(data)
    track = btrack.btypes.Tracklet(track_ID, objects)
    return track


def create_test_image(
    boxsize: int = 150,
    ndim: int = 2,
    nobj: int = 10,
    binsize: int = 5,
    *,
    binary: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """Make a test image that ensures that no two pixels are in contact."""

    rng = np.random.default_rng(seed=RANDOM_SEED)

    shape = (boxsize,) * ndim
    img = np.zeros(shape, dtype=np.uint16)

    # return an empty image if we have no objects
    if nobj == 0:
        return img, None

    # split this into voxels
    bins = boxsize // binsize

    def _sample() -> Tuple[np.ndarray, Tuple[int]]:
        _img = np.zeros((binsize,) * ndim, dtype=np.uint16)
        _coord = tuple(rng.integers(1, binsize - 1, size=(ndim,)).tolist())
        _img[_coord] = 1
        assert (
            np.sum(_img) == 1
        ), "Test image voxel contains incorrect number of objects."
        return _img, _coord

    # now we update nobj grid positions with a sample
    grid = np.stack(np.meshgrid(*[np.arange(bins)] * ndim), -1).reshape(
        -1, ndim
    )

    rbins = rng.choice(grid, size=(nobj,), replace=False)

    # iterate over the bins and add a smaple
    centroids = []
    for v, bin in enumerate(rbins):  # noqa: A001
        sample, point = _sample()
        slices = tuple(
            [slice(b * binsize, b * binsize + binsize, 1) for b in bin]
        )
        val = 1 if binary else v + 1
        img[slices] = sample * val

        # shift the actual coordinates back to image space
        point = point + bin * binsize
        centroids.append(point)

    # sort the centroids by axis
    centroids_sorted = np.array(centroids)
    centroids_sorted = centroids_sorted[
        np.lexsort([centroids_sorted[:, dim] for dim in range(ndim)][::-1])
    ]

    assert (
        centroids_sorted.shape[0] == nobj
    ), "Number of created centroids != requested in test image."

    vals = np.unique(img)
    assert (
        np.max(vals) == 1 if binary else nobj
    ), "Test image labels are incorrect."
    return img, centroids_sorted


def create_test_segmentation_and_tracks(
    boxsize: int = 128,
    padding: int = 16,
    nframes: int = 10,
    ndim: int = 2,
    *,
    binary: bool = False,
) -> Tuple[np.ndarray, np.ndarray, List[btrack.btypes.Tracklet]]:
    """Create a test segmentation with four tracks."""

    if ndim not in (btrack.constants.Dimensionality.TWO,):
        raise ValueError("Only 2D-segmentation currently supported.")

    # make a segmentation volume (10, 128, 128) for ndim == 2
    volume = tuple([nframes] + [boxsize] * ndim)
    segmentation = np.zeros(volume, dtype=np.int32)
    ground_truth = np.zeros_like(segmentation)

    dxy = (boxsize - 2 * padding) / nframes

    # create tracks moving from each corner of the segmentation at a constant
    # velocity towards another corner. these tracks do not intersect.
    track_A = create_realistic_tracklet(padding, padding, dxy, 0, nframes, 1)
    track_B = create_realistic_tracklet(
        boxsize - padding, boxsize - padding, -dxy, 0, nframes, 2
    )
    track_C = create_realistic_tracklet(
        padding, boxsize - padding, 0, -dxy, nframes, 3
    )
    track_D = create_realistic_tracklet(
        boxsize - padding, padding, 0, dxy, nframes, 4
    )

    tracks = [track_A, track_B, track_C, track_D]

    # set the segmentation values
    for track in tracks:
        t, y, x = np.split(
            track.to_array(properties=["t", "y", "x"]).astype(int), 3, 1
        )
        segmentation[t, y, x] = 1
        ground_truth[t, y, x] = track.ID

    if not binary:
        for n in range(nframes):
            segmentation[n, ...] = label(segmentation[n, ...])

    return segmentation, ground_truth, tracks


def full_tracker_example(
    objects: List[btrack.btypes.PyTrackObject], **kwargs
) -> btrack.BayesianTracker:
    """Set up a full tracker example. kwargs can supply configuration options."""
    # run the tracking
    tracker = btrack.BayesianTracker()
    tracker.configure(CONFIG_FILE)
    for cfg_key, cfg_value in kwargs.items():
        setattr(tracker, cfg_key, cfg_value)
    tracker.append(objects)
    tracker.volume = ((0, 1600), (0, 1200), (-1e5, 1e5))
    tracker.track(step_size=100)
    tracker.optimize()
    return tracker


def simple_tracker_example() -> Tuple[btrack.BayesianTracker, Dict[str, Any]]:
    """Run a simple tracker example with some data."""
    x = np.array([200, 201, 202, 203, 204, 207, 208])
    y = np.array([503, 507, 499, 500, 510, 515, 518])
    t = np.array([0, 1, 2, 3, 4, 5, 6])
    z = np.zeros(x.shape)

    objects_dict = {"x": x, "y": y, "z": z, "t": t}
    objects = btrack.io.objects_from_dict(objects_dict)

    tracker = full_tracker_example(objects)
    return tracker, objects_dict



import json

import numpy as np
import pytest

from btrack import models, utils

from ._utils import CONFIG_FILE


def test_read_motion_model():
    """Test reading the motion model."""

    with open(CONFIG_FILE, "r") as config_file:
        config = json.load(config_file)["TrackerConfig"]

    model = utils.read_motion_model(config)
    assert isinstance(model, models.MotionModel)


def test_incorrect_reshape_motion_model():
    """Test that specifying the wrong matrix dimensions raise a `ValueError`."""

    with open(CONFIG_FILE, "r") as config_file:
        config = json.load(config_file)["TrackerConfig"]

    m = config["MotionModel"]["measurements"]
    s = config["MotionModel"]["states"]

    with pytest.raises(ValueError):
        config["MotionModel"]["measurements"] = m + 1  # mess up shape
        _ = utils.read_motion_model(config)

    with pytest.raises(ValueError):
        config["MotionModel"]["measurements"] = m  # mess up shape
        config["MotionModel"]["states"] = s + 1  # mess up shape
        _ = utils.read_motion_model(config)


def test_read_hypothesis_model():
    """Test reading the hypothesis model."""

    with open(CONFIG_FILE, "r") as config_file:
        config = json.load(config_file)["TrackerConfig"]

    model = utils.read_hypothesis_model(config)
    assert isinstance(model, models.HypothesisModel)


def test_missing_process_covariance_motion_model():
    """Test missing process covariance in motion model."""
    with open(CONFIG_FILE, "r") as config_file:
        config = json.load(config_file)["TrackerConfig"]

    # delete the G matrix
    del config["MotionModel"]["G"]

    with pytest.raises(ValueError):
        _ = utils.read_motion_model(config)


def test_specifying_process_covariance_Q_motion_model():
    """Test specifying process covariance in motion model using Q."""
    with open(CONFIG_FILE, "r") as config_file:
        config = json.load(config_file)["TrackerConfig"]

    assert "G" in config["MotionModel"]
    assert "Q" not in config["MotionModel"]

    model = utils.read_motion_model(config)

    # delete the G matrix
    test_config = dict(config)
    sigma = test_config["MotionModel"]["G"]["sigma"]
    del test_config["MotionModel"]["G"]

    # create a test configuration specifying a Q matrix instead of a G matrix
    test_config["MotionModel"]["Q"] = {
        "sigma": sigma,
        "matrix": (model.Q.ravel() / sigma).tolist(),
    }

    assert "G" not in test_config["MotionModel"]
    assert "Q" in test_config["MotionModel"]

    test_model = utils.read_motion_model(test_config)

    np.testing.assert_equal(model.Q, test_model.Q)
